/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: descript.c,v 1.11 2021/05/29 10:27:42 babolo Exp $"

#define BLIN_COMPAT     4
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <mife.h>
#include "pngtr.h"

static const pngtr_chidef p_IEND =
{ 1, { { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_IEND = &p_IEND;

static const pngtr_chidef p_IHDR =
/*opn     op.descr                        maxl minl         var   otype  */
{ 8, { { "Width"                  ,          0,  4,           0, PNGTR_DU   }
     , { "Height"                 ,          0,  4,           0, PNGTR_DU   }
     , { "Bit depth"              ,          0,  1,           0, PNGTR_DU   }
     , { "Color type"             ,          0,  1,           0, PNGTR_DW1T }
     , { "Compression method"     ,          0,  1,           0, PNGTR_DU1N }
     , { "Filter method"          ,          0,  1,           0, PNGTR_DU1N }
     , { "Interlace method"       ,          0,  1,           0, PNGTR_DU1I }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_IHDR = &p_IHDR;

static const pngtr_chidef p_PLTE =
{ 2, { { "Palette"                ,          0,  3, PNGTR_OPIDX, PNGTR_DCC  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_PLTE = &p_PLTE;

static const pngtr_chidef p_IDAT =
{ 2, { { "Actual image data"      ,          0,  1, PNGTR_OPIDX, PNGTR_DBHZ }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_IDAT = &p_IDAT;

static const pngtr_chidef p_cHRM =
{ 9, { { "White Point x"          ,          0,  4,           0, PNGTR_DF4  }
     , { "White Point y"          ,          0,  4,           0, PNGTR_DF4  }
     , { "Red x"                  ,          0,  4,           0, PNGTR_DF4  }
     , { "Red y"                  ,          0,  4,           0, PNGTR_DF4  }
     , { "Green x"                ,          0,  4,           0, PNGTR_DF4  }
     , { "Green y"                ,          0,  4,           0, PNGTR_DF4  }
     , { "Blue x"                 ,          0,  4,           0, PNGTR_DF4  }
     , { "Blue y"                 ,          0,  4,           0, PNGTR_DF4  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_cHRM = &p_cHRM;

static const pngtr_chidef p_gAMA =
{ 2, { { "Gamma"                  ,          0,  4,           0, PNGTR_DF4  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_gAMA = &p_gAMA;

static const pngtr_chidef p_iCCP =
{ 4, { { "Profile name"           ,         80,  2, PNGTR_OPVAR, PNGTR_DCHAR} 
     , { "Compression method"     ,          0,  1,           0, PNGTR_DU1N }
     , { "ICC profile"            ,          0,  1, PNGTR_OPIDX, PNGTR_DBZ  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_iCCP = &p_iCCP;

static const pngtr_chidef p_sBIT0 =
{ 2, { { "Number of grey bits"    ,          0,  1,           0, PNGTR_DU   }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_sBIT2 =
{ 4, { { "Number of red bits"     ,          0,  1,           0, PNGTR_DU   }
     , { "Number of green bits"   ,          0,  1,           0, PNGTR_DU   }
     , { "Number of blue bits"    ,          0,  1,           0, PNGTR_DU   }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_sBIT4 =
{ 3, { { "Number of grey bits"    ,          0,  1,           0, PNGTR_DU   }
     , { "Number of alfa bits"    ,          0,  1,           0, PNGTR_DU   }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_sBIT6 =
{ 5, { { "Number of red bits"     ,          0,  1,           0, PNGTR_DU   }
     , { "Number of green bits"   ,          0,  1,           0, PNGTR_DU   }
     , { "Number of blue bits"    ,          0,  1,           0, PNGTR_DU   }
     , { "Number of alfa bits"    ,          0,  1,           0, PNGTR_DU   }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_sBIT[] =
{ &p_sBIT0, NULL, &p_sBIT2, &p_sBIT2, &p_sBIT4, NULL, &p_sBIT6, NULL };

static const pngtr_chidef p_sRGB =
{ 2, { { "Rendering intent"       ,          0,  1,           0, PNGTR_DW1R }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_sRGB = &p_sRGB;

static const pngtr_chidef p_bKGD0 =
{ 2, { { "Gray level value"       ,          0,  2,           0, PNGTR_DW2  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_bKGD2 =
{ 2, { { "Color levels value"     ,          0,  6,           0, PNGTR_DC2  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_bKGD3 =
{ 2, { { "Palette index"          ,          0,  1,           0, PNGTR_DW1  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef *pp_bKGD[] =
{ &p_bKGD0, NULL, &p_bKGD2, &p_bKGD3, &p_bKGD0, NULL, &p_bKGD2, NULL };

static const pngtr_chidef p_hIST =
{ 2, { { ""                       ,          0,  2, PNGTR_OPIDX, PNGTR_DUH  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_hIST = &p_hIST;

static const pngtr_chidef p_tRNS0 =
{ 2, { { "Gray level value"       ,          0,  2,           0, PNGTR_DW2  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_tRNS2 =
{ 2, { { "Color levels value"     ,          0,  6,           0, PNGTR_DC2  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_tRNS3 =
{ 2, { { "Alfa for palette"       ,          0,  1, PNGTR_OPIDX, PNGTR_DCA  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef *pp_tRNS[] =
{ &p_tRNS0, NULL, &p_tRNS2, &p_tRNS3, NULL, NULL, NULL, NULL };

static const pngtr_chidef p_pHYs =
{ 4, { { "Pixels per unit, X axis",          0,  4,           0, PNGTR_DU   }
     , { "Pixels per unit, Y axis",          0,  4,           0, PNGTR_DU   }
     , { "Unit specifier"         ,          0,  1,           0, PNGTR_DUP  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_pHYs = &p_pHYs;

static const pngtr_chidef p_sPLT0 =
{ 4, { { "Palette name"           ,         80,  2, PNGTR_OPVAR, PNGTR_DCHAR} 
     , { "Sample depth"           ,          0,  1, PNGTR_CHDEP, PNGTR_DU1W }
     , { "Recommended palette"    ,          0,  6, PNGTR_OPIDX, PNGTR_DCA1 }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_sPLT1 =
{ 4, { { "Palette name"           ,         80,  2, PNGTR_OPVAR, PNGTR_DCHAR} 
     , { "Sample depth"           ,          0,  1, PNGTR_CHDEP, PNGTR_DU1W }
     , { "Recommended palette"    ,          0, 10, PNGTR_OPIDX, PNGTR_DCA2 }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_sPLT[] = { &p_sPLT0, &p_sPLT1 };

static const pngtr_chidef p_tIME =
{ 7, { { "Year"                   ,          0,  2,           0, PNGTR_DU   }
     , { "Month"                  ,          0,  1,           0, PNGTR_DU   }
     , { "Day"                    ,          0,  1,           0, PNGTR_DU   }
     , { "Hour"                   ,          0,  1,           0, PNGTR_DU   }
     , { "Minute"                 ,          0,  1,           0, PNGTR_DU   }
     , { "Second"                 ,          0,  1,           0, PNGTR_DU   }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_tIME = &p_tIME;

static const pngtr_chidef p_iTXt0 =
{ 7, { { "Keyword"                ,         80,  2, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Compression flag"       ,          0,  1, PNGTR_CHZIP, PNGTR_DU1C }
     , { "Compression method"     ,          0,  1,           0, PNGTR_DU1N }
     , { "Language tag"           , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Translated keyword"     , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Text"                   , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };

static const pngtr_chidef p_iTXt1 =
{ 7, { { "Keyword"                ,         80,  2, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Compression flag"       ,          0,  1, PNGTR_CHZIP, PNGTR_DU1C }
     , { "Compression method"     ,          0,  1,           0, PNGTR_DU1N }
     , { "Language tag"           , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Translated keyword"     , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Text"                   , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAZ}
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_iTXt[] = { &p_iTXt0, &p_iTXt1 };

static const pngtr_chidef p_tEXt =
{ 3, { { "Keyword"                ,         80,  2, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Text"                   , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_tEXt = &p_tEXt;

static const pngtr_chidef p_zTXt =
{ 4, { { "Keyword"                ,         80,  2, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Compression method"     ,          0,  1,           0, PNGTR_DU1N }
     , { "Text"                   , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAZ}
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_zTXt = &p_zTXt;

static const pngtr_chidef p_oFFs =
{ 4, { { "X position"             ,          0,  4,           0, PNGTR_DD4  }
     , { "Y position"             ,          0,  4,           0, PNGTR_DD4  }
     , { "Unit specifier"         ,          0,  1,           0, PNGTR_DUS  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_oFFs = &p_oFFs;

static const pngtr_chidef p_pCAL =
{11, { { "Calibration name"       ,         80,  2, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Original zero"          ,          0,  4,           0, PNGTR_DD4  }
     , { "Original max"           ,          0,  4,           0, PNGTR_DD4  }
     , { "Equation type"          ,          0,  1,           0, PNGTR_DU   }
     , { "Number of parameters"   ,          0,  1, PNGTR_OPNUM, PNGTR_DU   }
     , { "Unit name"              , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Parameter 0"            , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Parameter 1"            , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Parameter 2"            , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { "Parameter 3"            , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_pCAL = &p_pCAL;

static const pngtr_chidef p_sCAL =
{ 4, { { "Unit specifier"         ,          0,  1,           0, PNGTR_DUC  }
     , { "Pixel width"            , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DTFIX}
     , { "Pixel height"           , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DTFIX}
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_sCAL = &p_sCAL;

static const pngtr_chidef p_gIFg =
{ 4, { { "Disposal Method"        ,          0,  1,           0, PNGTR_DW1  }
     , { "User Input Flag"        ,          0,  1,           0, PNGTR_DW1  }
     , { "Delay Time"             ,          0,  2,           0, PNGTR_DF2  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_gIFg = &p_gIFg;

static const pngtr_chidef p_gIFt =
{10, { { "Text Grid Left Position",          0,  4,           0, PNGTR_DW4  }
     , { "Text Grid Top Position" ,          0,  4,           0, PNGTR_DW4  }
     , { "Text Grid Width"        ,          0,  4,           0, PNGTR_DW4  }
     , { "Text Grid Height"       ,          0,  4,           0, PNGTR_DW4  }
     , { "Character Cell Width"   ,          0,  1,           0, PNGTR_DW1  }
     , { "Character Cell Height"  ,          0,  1,           0, PNGTR_DW1  }
     , { "Text Foreground Color"  ,          0,  3,           0, PNGTR_DC1  }
     , { "Text Background Color"  ,          0,  3,           0, PNGTR_DC1  }
     , { "Plain Text Data"        , 0xFFFFFFFF,  1, PNGTR_OPVAR, PNGTR_DCHAR}
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_gIFt = &p_gIFt;

static const pngtr_chidef p_gIFx =
{ 4, { { "Application Identifier" ,          8,  1, PNGTR_OPFIX, PNGTR_DCHAR}
     , { "Authentication Code"    ,          3,  1, PNGTR_OPFIX, PNGTR_DBYTE}
     , { "Application Data"       ,          0,  1, PNGTR_OPIDX, PNGTR_DBG  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_gIFx = &p_gIFx;

static const pngtr_chidef p_sTER =
{ 2, { { "Layout"                 ,          0,  1,           0, PNGTR_DW1S }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_sTER = &p_sTER;

static const pngtr_chidef p_eXIf =
{ 3, { { "Signature"              ,          4,  1, PNGTR_OPFIX, PNGTR_DBYTE}
     , { "Data"                   ,          0,  1, PNGTR_OPIDX, PNGTR_DBG  }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_eXIf = &p_eXIf;

static const pngtr_chidef p_acTL =
{ 3, { { "num_frames"             ,          0,  4,           0, PNGTR_DU   }
     , { "num_plays"              ,          0,  4,           0, PNGTR_DU   }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_acTL = &p_acTL;

static const pngtr_chidef p_fcTL =
{10, { { "Sequence Number"           ,       0,  4,           0, PNGTR_DU   }
     , { "Width of the frame"        ,       0,  4,           0, PNGTR_DU   }
     , { "Height of the frame"       ,       0,  4,           0, PNGTR_DU   }
     , { "X offset of the frame"     ,       0,  4,           0, PNGTR_DU   }
     , { "Y offset of the frame"     ,       0,  4,           0, PNGTR_DU   }
     , { "Delay fraction numerator"  ,       0,  2,           0, PNGTR_DU   }
     , { "Delay fraction denominator",       0,  2,           0, PNGTR_DUNO }
     , { "Dispose operation after rendering",0,  1,           0, PNGTR_DUOD }
     , { "Blend operation"           ,       0,  1,           0, PNGTR_DUOB }
     , { ""                          ,       0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_fcTL = &p_fcTL;

static const pngtr_chidef p_fdAT =
{ 3, { { "Sequence Number"        ,          0,  4,           0, PNGTR_DU   }
     , { "Actual frame data"      ,          0,  1, PNGTR_OPIDX, PNGTR_DBHZ }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_fdAT = &p_fdAT;

static const pngtr_chidef p_vpAg =
{ 4, { { "Width"                  ,          0,  4,           0, PNGTR_DU   }
     , { "Height"                 ,          0,  4,           0, PNGTR_DU   }
     , { "Units"                  ,          0,  1,           0, PNGTR_DU   }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_vpAg = &p_vpAg;

static const pngtr_chidef p_JHDR =
{11, { { "Width"                   ,         0,  4,           0, PNGTR_DU   }
     , { "Height"                  ,         0,  4,           0, PNGTR_DU   }
     , { "Color type"              ,         0,  1,           0, PNGTR_DW1T }
     , { "Image sample depth"      ,         0,  1,           0, PNGTR_DU   }
     , { "Image compression method",         0,  1,           0, PNGTR_DU   }
     , { "Image interlace method"  ,         0,  1,           0, PNGTR_DU   }
     , { "Alpha sample depth"      ,         0,  1,           0, PNGTR_DU   }
     , { "Alpha compression method",         0,  1,           0, PNGTR_DU   }
     , { "Alpha filter method"     ,         0,  1,           0, PNGTR_DU1N }
     , { "Alpha interlace method"  ,         0,  1,           0, PNGTR_DU1N }
     , { ""                        ,         0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_JHDR = &p_JHDR;

static const pngtr_chidef p_JDAT =
{ 2, { { "Actual image data"      ,          0,  1, PNGTR_OPIDX, PNGTR_DBHZ }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_JDAT = &p_JDAT;

static const pngtr_chidef p_JDAA =
{ 2, { { "Actual alfa data"       ,          0,  1, PNGTR_OPIDX, PNGTR_DBHZ }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_JDAA = &p_JDAA;

static const pngtr_chidef p_dSIG =
{ 2, { { "Digital Signature"      ,          0,  1, PNGTR_OPIDX, PNGTR_DBHZ }
     , { ""                       ,          0,  1, PNGTR_OPIDX, PNGTR_TRASH}
}    };
static const pngtr_chidef *pp_dSIG = &p_dSIG;

const pngtr_chdef pngtr_chdescr[] =
/*  descr                             nm              shift       ch */
{ {                             NULL, "????",             0, &pp_IEND }
, {"Header"                         , "IHDR", PNGTR_PNGHEAD, &pp_IHDR }
, {"Palette"                        , "PLTE",             0, &pp_PLTE }
, {"Image data"                     , "IDAT",             0, &pp_IDAT }
, {"Image trailer"                  , "IEND",             0, &pp_IEND }
, {"Primary chromaticities"         , "cHRM",             0, &pp_cHRM }
, {"Color space information"        , "gAMA",             0, &pp_gAMA }
, {"Embedded ICC profile"           , "iCCP",             0, &pp_iCCP }
, {"Significant bits"               , "sBIT", PNGTR_CHCOLOR,  pp_sBIT }
, {"Standard RGB color space"       , "sRGB",             0, &pp_sRGB }
, {"Background color"               , "bKGD", PNGTR_CHCOLOR,  pp_bKGD }
, {"Palette histogram"              , "hIST",             0, &pp_hIST }
, {"Transparency"                   , "tRNS", PNGTR_CHCOLOR,  pp_tRNS } 
, {"Physical pixel dimensions"      , "pHYs",             0, &pp_pHYs }
, {"Suggested palette"              , "sPLT", PNGTR_CHDEP  ,  pp_sPLT }
, {"Last-modification time"         , "tIME",             0, &pp_tIME }
, {"International textual data"     , "iTXt", PNGTR_CHZIP  ,  pp_iTXt }
, {"Textual data"                   , "tEXt",             0, &pp_tEXt }
, {"Compressed textual data"        , "zTXt",             0, &pp_zTXt }
, {"Image offset"                   , "oFFs",             0, &pp_oFFs }
, {"Calibration of pixel values"    , "pCAL", PNGTR_OPNUM  , &pp_pCAL }
, {"Physical scale of image subject", "sCAL",             0, &pp_sCAL }
, {"GIF Graphic Control Extension"  , "gIFg",             0, &pp_gIFg }
, {"GIF Plain Text Extension"       , "gIFt",             0, &pp_gIFt }
, {"GIF Application Extension"      , "gIFx",             0, &pp_gIFx }
, {"Indicator of Stereo Image"      , "sTER",             0, &pp_sTER }
, {"Digital signature"              , "dSIG",             0, &pp_dSIG }
, {"Exchangeable Image File Profile", "eXIf",             0, &pp_eXIf }
, {"Fractal image parameters"       , "fRAc",             0, &pp_IEND }
, {"APNG Animation Control"         , "acTL",             0, &pp_acTL }
, {"APNG Frame Control"             , "fcTL",             0, &pp_fcTL }
, {"APNG Frame Data"                , "fdAT",             0, &pp_fdAT }
, {"ImageMagick virtual page"       , "vpAg",             0, &pp_vpAg }
, {"ImageMagick orientation"        , "orNT",             0, &pp_IEND }
, {"ImageMagick canvas"             , "caNv",             0, &pp_IEND }
, {"ImageMagick exIf read-write"    , "exIf",             0, &pp_IEND }
, {"JNG header"                     , "JHDR",             0, &pp_JHDR }
, {"JNG image data"                 , "JDAT",             0, &pp_JDAT }
, {"JNG JPEG-encoded alpha data"    , "JDAA",             0, &pp_JDAA }
, {"8-bit/12-bit image separator"   , "JSEP",             0, &pp_IEND }
};
