/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: print.c,v 1.7 2021/05/29 10:27:42 babolo Exp $"

#define BLIN_COMPAT     4
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <mife.h>
#include "pngtr.h"

static void
prtl(int l, FILE *pr) {
    static const char b[] = ".  .  .  .  .  .  .  .  .  .  . ";

    fprintf(pr, " %s", &b[(l < 31) ? l : 31]);
}
int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     pr   chctr  cf->num    f                ****
 *****************************************************************************************************
                                                                                                  ****/
chprint(pngtr_conf *cf, u_int chctr, FILE *pr) {                                                 /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey       *convey = &cf->convey[cf->num];
    pngtr_chunk        *chunk  = &convey->chunks[chctr];
    const pngtr_chdef  *chdef;
    u_int32_t           crc    = 1;
    pngtr_hdr          *hdr;
    int                 ex     = EX_OK;
    u_int32_t           hl;
    u_int8_t           *c      = convey->png.cont;
    int                 l;

    ifBLIN_QX3("+ %08X %08X q%u c%u", cf->flags, convey->flags, cf->num, chctr);
    hdr = (pngtr_hdr *)&c[chunk->offset];
    ifBLIN_QX4( "%"BLIN_X" %"BLIN_X" %u %08X"
              , BLIN_I(hdr)
              , BLIN_I(chunk->offset)
              , chunk->csz
              , chunk->flags
              );
    if  (10 > chctr) {
        fprintf(pr, "oooo%1u %08X", chctr, (u_int32_t)chunk->offset);
    } else if (100 > chctr) {
        fprintf(pr, "ooo%2u %08X", chctr, (u_int32_t)chunk->offset);
    } else if (1000 > chctr) {
        fprintf(pr, "oo%3u %08X", chctr, (u_int32_t)chunk->offset);
    } else if (10000 > chctr) {
        fprintf(pr, "o%4u %08X", chctr, (u_int32_t)chunk->offset);
    } else {
        fprintf(pr, "%5u %08X", chctr, (u_int32_t)chunk->offset);
    }
    if  (4 >= chunk->csz) {
        fprintf(pr, " **%08X**\n", chunk->csz);
        goto out;
    }
    hl = recobe_load(hdr->ln, -4);
    fprintf(pr, " %08X", hl);
    if  (12 > chunk->csz) {
        fprintf( pr
               , "%s -- %.*s<%s%s%s%s>\n"
               , (12 >= chunk->csz) ? "**" : ""
               , (chunk->csz < 8) ? chunk->csz - 4 : 4, hdr->nm
               , !(hdr->nm[0] & 0x20) ? "CRIT" : "anc"
               , ((5 < chunk->csz) && !(hdr->nm[1] & 0x20)) ? " pub" : " priv"
               , ((6 < chunk->csz) && !(hdr->nm[2] & 0x20)) ? "" : "!"
               , ((7 < chunk->csz) && !(hdr->nm[3] & 0x20)) ? "" : " copy"
               );
        goto out;
    }
    ifBLIN_QX5("%u %u", hl, chunk->csz);
    if  ((12 + hl) <= chunk->csz) {
        ifBLIN_QX5( "%08X %08X"
                  , (u_int32_t)recobe_load(CRC(hdr), -4)
                  , recobe_crc32((u_char *)hdr->nm, hl + 4)
                  );
        crc = recobe_load(CRC(hdr), -4) ^ recobe_crc32((u_char *)hdr->nm, hl + 4);
    }
    chdef = &pngtr_chdescr[chunk->type];
    l = fprintf( pr
               , " %s %.*s<%s%s%s%s>"
               , (PNGTR_VIRTUAL & chunk->flags) ? "" : !crc ? "OK" : "crc"
               , 4, chdef->nm
               , !(chdef->nm[0] & 0x20) ? "CRIT" : "anc"
               , !(chdef->nm[1] & 0x20) ? " pub" : " priv"
               , !(chdef->nm[2] & 0x20) ? "" : "!"
               , !(chdef->nm[3] & 0x20) ? "" : " copy"
               );
    if  ((PNGTR_IDAT == chunk->type) && (!chunk->inflen)) l += fprintf(pr, "!");
    if  (!chdef->descr) {
        prtl(l + 6, pr);
        fprintf(pr, "%.*s", 4, hdr->nm);
    } else if ((PNGTR_OPNUM & chdef->shift) && (6 < chunk->carg)){
        switch(chunk->operands[3].val[0]) {
        case 0:
            fprintf( pr
                   , " R = %.*s + %.*s * SAMPLE / (%d - %d)"
                   , chunk->operands[6].size, chunk->operands[6].val
                   , (7 < chunk->carg) ? chunk->operands[7].size : 1
                   , (7 < chunk->carg) ? (char *)chunk->operands[7].val : "?"
                   , (int32_t)recobe_load(chunk->operands[2].val, -4)
                   , (int32_t)recobe_load(chunk->operands[1].val, -4)
                   );
            break;
        case 1:
            fprintf( pr
                   , " R = %.*s + %.*s * exp(%.*s * SAMPLE / (%d - %d))"
                   , chunk->operands[6].size, chunk->operands[6].val
                   , (7 < chunk->carg) ? chunk->operands[7].size : 1
                   , (7 < chunk->carg) ? (char *)chunk->operands[7].val : "?"
                   , (8 < chunk->carg) ? chunk->operands[8].size : 1
                   , (8 < chunk->carg) ? (char *)chunk->operands[8].val : "?"
                   , (int32_t)recobe_load(chunk->operands[2].val, -4)
                   , (int32_t)recobe_load(chunk->operands[1].val, -4)
                   );
            break;
        case 2:
            fprintf( pr
                   , " R = %.*s + %.*s * pow(%.*s, (SAMPLE / (%d - %d)))"
                   , chunk->operands[6].size, chunk->operands[6].val
                   , (7 < chunk->carg) ? chunk->operands[7].size : 1
                   , (7 < chunk->carg) ? (char *)chunk->operands[7].val : "?"
                   , (8 < chunk->carg) ? chunk->operands[8].size : 1
                   , (8 < chunk->carg) ? (char *)chunk->operands[8].val : "?"
                   , (int32_t)recobe_load(chunk->operands[2].val, -4)
                   , (int32_t)recobe_load(chunk->operands[1].val, -4)
                   );
            break;
        case 3:
            fprintf( pr
                   , " R = %.*s + %.*s * sinh(%.*s * (SAMPLE - %.*s) / (%d - %d))"
                   , chunk->operands[6].size, chunk->operands[6].val
                   , (7 < chunk->carg) ? chunk->operands[7].size : 1
                   , (7 < chunk->carg) ? (char *)chunk->operands[7].val : "?"
                   , (8 < chunk->carg) ? chunk->operands[8].size : 1
                   , (8 < chunk->carg) ? (char *)chunk->operands[8].val : "?"
                   , (9 < chunk->carg) ? chunk->operands[9].size : 1
                   , (9 < chunk->carg) ? (char *)chunk->operands[9].val : "?"
                   , (int32_t)recobe_load(chunk->operands[2].val, -4)
                   , (int32_t)recobe_load(chunk->operands[1].val, -4)
                   );
            break;
        default:
            prtl(l + 6, pr);
            fprintf(pr, "%s", chdef->descr);
        }
    } else {
        prtl(l + 6, pr);
        fprintf(pr, "%s", chdef->descr);
    }
    fprintf(pr, "\n");
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     pr opctr   chctr  cf->num   f               ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_oprint(pngtr_conf *cf, u_int chctr, u_int opctr, BLIN_flag f, FILE *pr) {                  /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey       *convey = &cf->convey[cf->num];
    pngtr_chunk        *chunk  = &convey->chunks[chctr];
    const pngtr_chdef  *chdef;
    const pngtr_chidef *opdef;
    u_int               opnum;
    int                 ex     = EX_OK;
    pngtr_operand      *op     = &chunk->operands[opctr];
    int                 l      = 0;

    ifBLIN_QX3("+ %08X f%08X %08X q%u c%u o%u", cf->flags, f, convey->flags, cf->num, chctr, opctr);
    chdef = &pngtr_chdescr[chunk->type];
    opdef = chdef->ch[(PNGTR_CHCOLOR & chdef->shift) ? convey->IHDR.colortype : 0];
    opnum = opdef->opn;
    if  (!op->val) goto out;
    if  ((PNGTR_OPERSEL & convey->flags) && !(PNGTR_SELECTD & op->flags)) goto out;
    if  (  (1 < opctr)
        && (  ((PNGTR_CHDEP & opdef->op[1].var) && (op->val[0] > 8))
           || ((PNGTR_CHZIP & opdef->op[1].var) && !!op->val[0])
        )  ) {                                /*  sPLT0  iTXt     */
        opdef = chdef->ch[1];
    }
    ifBLIN_QX4("%08X [%u] %u x %u", op->flags, opdef->op[opctr].otype, op->size, op->maxindex);
    if  (PNGTR_PRINTOP & f) {
        if  (  (PNGTR_TRASH == opdef->op[opctr].otype)
            && (PNGTR_AGRERR & convey->flags)
            && (4 == op->size)
            ) {
            fprintf(pr, "       [CRC] ");
        } else if (PNGTR_TRASH == opdef->op[opctr].otype) {
            fprintf(pr, "     [Trash]");
        } else if (10 > opctr) {
            fprintf(pr, "        [%1u] ", opctr);
        } else if (100 > opctr) {
            fprintf(pr, "       [%2u] ", opctr);
        } else {
            fprintf(pr, "      [%3u] ", opctr);
        }
        switch(opdef->op[opctr].otype) {
        case PNGTR_DU:
            l = fprintf(pr, "%u", (u_int32_t)recobe_load(op->val, -(int)op->size));
            break;
        case PNGTR_DD4:
            l = fprintf(pr, "%d", (u_int32_t)recobe_load(op->val, -(int)op->size));
            break;
        case PNGTR_DF4:
            l = fprintf(pr, "%f", (u_int32_t)recobe_load(op->val, -(int)op->size) / 100000.0);
            break;
        case PNGTR_DF2:
            l = fprintf(pr, "%f", (u_int32_t)recobe_load(op->val, -(int)op->size) / 100.0);
            break;
        case PNGTR_DW2:
            l = fprintf(pr, "0x%04X", (u_int32_t)recobe_load(op->val, -(int)op->size));
            break;
        case PNGTR_DW1:
            l = fprintf(pr, "0x%02X", *(op->val));
            break;
        case PNGTR_DCHAZ:
            if  ((!chunk->infla) && !!(ex = pngtr_infla(cf, chctr, opctr))) {
                l = fprintf(pr, "Error");
                ifBLIN_QW0( "infla %s %u %u"
                          , !convey->png.nm ? "" : convey->png.nm
                          , chctr
                          , opctr
                          );
                break;
            }
            l = fprintf(pr, "|%.*s|", (int)chunk->inflen, chunk->infla);
            goto dchar;
        case PNGTR_DCHAR:
        case PNGTR_DTFIX:
            l = fprintf(pr, "|%.*s|", (int)op->size, op->val);
        dchar:
#if 0
            for (u_int32_t i = 0; i < op->size; ++i) {
                if  (' ' > op->val[i]) {
                    l = -1;
                    break;
            }   }
#endif
            if  (31 < l) {
                fprintf(pr, "\n");
                l = -1;
            }
            break;
        case PNGTR_DW1S: {
                const static char *y[] = {"cross-fuse", "diverging-fuse", "Illegal"};

                l = fprintf(pr, "0x%02X %s", *(op->val), y[(*(op->val) > 2) ? 2 : *(op->val)]);
            }
            break;
        case PNGTR_DW1R: {
                const static char *y[] = { "Perceptual"
                                         , "Relative colorimetric"
                                         , "Saturation"
                                         , "Absolute colorimetric"
                                         , "Illegal"
                                         };

                l = fprintf(pr, "0x%02X %s", *(op->val), y[(*(op->val) > 4) ? 4 : *(op->val)]);
            }
            break;
        case PNGTR_DU1W:
            l = fprintf( pr
                       , "%u %s"
                       , *(op->val)
                       , ((8 != *(op->val)) && (16 != *(op->val))) ? "Illegal" : ""
                       );
            break;
        case PNGTR_DU1C: {
                const static char *y[] = {"Not compressed", "Compressed", "Illegal" };

                l = fprintf(pr, "%u %s", *(op->val), y[(*(op->val) > 2) ? 2 : *(op->val)]);
            }
            break;
        case PNGTR_DUS: {
                const static char *y[] = {"pixel", "micrometer", "Illegal"};

                l = fprintf(pr, "%u %s", *(op->val), y[(*(op->val) > 2) ? 2 : *(op->val)]);
            }
            break;
        case PNGTR_DUC: {
                const static char *y[] = {"unknown", "meter", "radian", "Illegal"};

                l = fprintf(pr, "%u %s", *(op->val), y[(*(op->val) > 3) ? 3 : *(op->val)]);
            }
            break;
        case PNGTR_DUP: {
                const static char *y[] = {"unknown", "meter", "Illegal"};

                l = fprintf(pr, "%u %s", *(op->val), y[(*(op->val) > 2) ? 2 : *(op->val)]);
            }
            break;
        case PNGTR_DU1N:
            l = fprintf(pr, "%u %s", *(op->val), !!*(op->val) ? " Illegal" : "");
            break;
        case PNGTR_DUOD: {
                const static char *y[] = {"None", "Background", "Previous", "Illegal"};

                l = fprintf(pr, "%u %s", *(op->val), y[(*(op->val) > 3) ? 3 : *(op->val)]);
            }
            break;
        case PNGTR_DUOB: {
                const static char *y[] = {"Source", "Over", "Illegal"};

                l = fprintf(pr, "%u %s", *(op->val), y[(*(op->val) > 2) ? 2 : *(op->val)]);
            }
            break;
        case PNGTR_DUNO:
            if  (!*(op->val)) l = fprintf(pr, "0 (/100)"); else l = fprintf(pr, "/%u", *(op->val));
            break;
        case PNGTR_DW1T: {
                pngtr_hdrcolor v;
                pngtr_operand *om = &chunk->operands[opctr - 1];
                int            x = 0;

                v.colortype = *(op->val);
                if  (16 < *(om->val)) ++x;
                if  (14 < *(op->val)) ++x;
                if  (!x && !pngtr_pointsize[*(op->val)][*(om->val)]) ++x;
                l = fprintf( pr
                           , "<%02X%s%s%s%s%s>%s"
                           , v.colortype
                           , !(v.palette) ? "" : " palette"
                           , !(v.color)   ? "" : " color"
                           , !(v.alfa)    ? "" : " alfa"
                           , !(v.j)       ? "" : " jpg"
                           , !(v.q)       ? "" : " ?"
                           , !x           ? "" : " Illegal"
                           );
            }
            break;
        case PNGTR_DU1I: {
                const static char *y[] = {"", "Adam7", "Illegal"};

                l = fprintf(pr, "%u %s", *(op->val), y[(*(op->val) > 2) ? 2 : *(op->val)]);
            }
            break;
        case PNGTR_DC1: {
                pngtr_color *t = (void*)op->val;

                l = fprintf(pr, "r#%02X g#%02X b#%02X", t->r, t->g, t->b);
            }
            break;
        case PNGTR_DC2: {
                pngtr_color2 *t = (void*)op->val;

                l = fprintf( pr
                           , "r#%04X g#%04X b#%04X"
                           , (u_int16_t)recobe_load(t->r, -2)
                           , (u_int16_t)recobe_load(t->g, -2)
                           , (u_int16_t)recobe_load(t->b, -2)
                           );
            }
            break;
        case PNGTR_DBYTE:
            for (u_int32_t i = 0; i < op->maxindex; ++i) l += fprintf(pr, "%02X", op->val[i]);
            break;
        case PNGTR_TRASH:
            if  (!(PNGTR_AGRERR & convey->flags)) {
                fprintf(pr, " **** %u + 4 Bytes ****\n", op->maxindex);
            } else {
                fprintf(pr, " **** %u Bytes ****\n", op->maxindex);
            }
            l = -1;
            break;
        case PNGTR_DCC:
        case PNGTR_DCA:
        case PNGTR_DUH:
            l = -1;
            break;
        }
        if  (0 <= l) {
            prtl(l, pr);
            fprintf(pr, "%s\n", opdef->op[opctr].descr);
        }
        switch(opdef->op[opctr].otype) {
        case PNGTR_DCC: {
                pngtr_color  *n = (void*)op->val;
                pngtr_miaobj *m;

                PNGTR_W(!(m = pngtr_minit(cf, 3, 1)),                     ("minit 3 1"), EX_OSERR);
                fprintf( pr
                       , " RGB   0       1       2       3       4       5       6       7"
                       );
                for (u_int32_t i = 0; i < op->maxindex; ++i) {
                    if  (!(i & 15)) {
                        fprintf(pr, "\n            %03X ", i >> 4);
                    } else if (!(i & 7)) {
                        fprintf(pr, "\n                ");
                    }
                    fprintf(pr, " #%02X%02X%02X", n[i].r, n[i].g, n[i].b);
                    PNGTR_O(!!(ex = pngtr_minimax(cf, m, &n[i])),                ("minmax %u", i));
                }
                if  (!!op->maxindex) {
                    fprintf( pr
                           , "\n                       min #%02X%02X%02X     max #%02X%02X%02X"
                           , m->minimax[0].min
                           , m->minimax[1].min
                           , m->minimax[2].min
                           , m->minimax[0].max
                           , m->minimax[1].max
                           , m->minimax[2].max
                           );
                }
                if  (256 < op->maxindex) {
                    fprintf(pr, "   **** VERY LONG PALETTE: %u ****", op->maxindex);
                }
                free(m);
                fprintf(pr, "\n");
            }
            break;
        case PNGTR_DCA: {
                pngtr_miaobj *m;

                PNGTR_W(!(m = pngtr_minit(cf, 1, 1)),                          ("minit 1 1"), EX_OSERR);
                fprintf( pr
                       , "  alfa 0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F"
                       );
                for (u_int32_t i = 0; i < op->maxindex; ++i) {
                    if  (!(i & 15)) fprintf(pr, "\n            %03X ", i >> 4);
                    fprintf(pr, " #%02X", op->val[i]);
                    PNGTR_O(!!(ex = pngtr_minimax(cf, m, &op->val[i])),               ("minmax %u", i));
                }
                if  (!!op->maxindex) fprintf( pr
                                           , "\n                   min #%02X     max #%02X"
                                           , m->minimax[0].min
                                           , m->minimax[0].max
                                           );
                if  (256 < op->maxindex) {
                    fprintf(pr, "**** VERY LONG ALFA: %u ****", op->maxindex);
                }
                free(m);
                fprintf(pr, "\n");
            }
            break;
        case PNGTR_DCA1: {
                pngtr_miaobj  *mc;
                pngtr_miaobj  *mf;
                u_int16_t      f;
                pngtr_sPLT1_t *t;

                PNGTR_W(!(mf = pngtr_minit(cf, 1, 2)),                        ("minit 1 2"), EX_OSERR);
                PNGTR_W(!(mc = pngtr_minit(cf, 4, 1)),                        ("minit 4 1"), EX_OSERR);
                t = (void*)op->val; 
                for (u_int32_t i = 0; i < op->maxindex; ++i, ++t) {
                    f = recobe_load(t->freq, -2);
                    fprintf( pr
                           , "           r#%02X g#%02X b#%02X a#%02X %5u"
                           , t->color.r
                           , t->color.g
                           , t->color.b
                           , t->a
                           , f
                           );
                    PNGTR_O(!!(ex = pngtr_minimax(cf, mf, &f)),                            ("minmax"));
                    PNGTR_O(!!(ex = pngtr_minimax(cf, mc, &t->color)),                     ("minmax"));
                }
                if  (!!op->maxindex) fprintf( pr
                                            , "\n         min r#%02X g#%02X b#%02X a#%02X"
                                              " max r#%02X g#%02X b#%02X a#%02X %5u"
                                            , mc->minimax[0].min
                                            , mc->minimax[1].min
                                            , mc->minimax[2].min
                                            , mc->minimax[3].min
                                            , mc->minimax[0].max
                                            , mc->minimax[1].max
                                            , mc->minimax[2].max
                                            , mc->minimax[3].max
                                            , mf->minimax[0].max
                                            );
                if  (256 < op->maxindex) {
                    fprintf(pr, "   **** VERY LONG PALETTE: %u ****", op->maxindex);
                }
                free(mf);
                free(mc);
                if  (!!op->maxindex) fprintf(pr, "\n");
            }
            break;
        case PNGTR_DCA2: {
                pngtr_miaobj  *mc;
                pngtr_sPLT2_t *t;

                PNGTR_W(!(mc = pngtr_minit(cf, 5, -2)),                  ("minit 5 -2"), EX_OSERR);
                t = (void*)op->val;
                for (u_int32_t i = 0; i < op->maxindex; ++i, ++t) {
                    fprintf( pr
                           , "\n           r#%04X g#%04X b#%04X a#%04X %5u"
                           , (u_int16_t)recobe_load(t->color.r, -2)
                           , (u_int16_t)recobe_load(t->color.g, -2)
                           , (u_int16_t)recobe_load(t->color.b, -2)
                           , (u_int16_t)recobe_load(t->a, -2)
                           , (u_int16_t)recobe_load(t->freq, -2)
                           );
                    PNGTR_O(!!(ex = pngtr_minimax(cf, mc, &t)),                        ("minmax"));
                }
                if  (!!op->maxindex) fprintf( pr
                                           , "\n          min r#%04X g#%04X b#%04X a#%04X"
                                             " max r#%04X g#%04X b#%04X a#%04X %5u"
                                           , mc->minimax[0].min
                                           , mc->minimax[1].min
                                           , mc->minimax[2].min
                                           , mc->minimax[3].min
                                           , mc->minimax[0].max
                                           , mc->minimax[1].max
                                           , mc->minimax[2].max
                                           , mc->minimax[3].max
                                           , mc->minimax[4].max
                                           );
                if  (256 < op->maxindex) {
                    fprintf(pr, "   **** VERY LONG PALETTE: %u ****", op->maxindex);
                }
                free(mc);
                if  (!!op->maxindex) fprintf(pr, "\n");
            }
            break;
        case PNGTR_DUH: {
                u_int16_t    *t = (void*)op->val;
                pngtr_miaobj *m;
                u_int16_t     f;

                PNGTR_W(!(m = pngtr_minit(cf, 1, 2)),                     ("minit 1 2"), EX_OSERR);
                fprintf(pr, " hist    0     1     2     3     4     5     6     7");
                for (u_int32_t i = 0; i < op->maxindex; ++i, ++t) {
                    f = recobe_load((void *)t, -2);
                    if  (!(i & 15)) {
                        fprintf(pr, "\n            %03X ", i >> 4);
                    } else if (!(i & 7)) {
                        fprintf(pr, "\n                ");
                    }
                    fprintf(pr, " %5u", f);
                    PNGTR_O(!!(ex = pngtr_minimax(cf, m, &f)),                   ("minmax %u", i));
                }
                if  (!!op->maxindex) {
                    fprintf(pr, "\n                       max %5u", m->minimax[0].max);
                }
                if  (256 < op->maxindex) {
                    fprintf(pr, "   **** VERY LONG PALETTE: %u ****", op->maxindex);
                }
                free(m);
                fprintf(pr, "\n");
            }
            break;
    }   }
    if  (PNGTR_PRINTXD & f) {
        switch(opdef->op[opctr].otype) {
        case PNGTR_TRASH:
            PNGTR_O( !!(ex = pngtr_dump( cf
                                       , op->val
                                       ,   op->size * op->maxindex
                                         + (!(PNGTR_AGRERR & convey->flags) ? 4 : 0)
                                       , pr
                       )               )
                   , ( "pngtr_dump %s q%u c%u o%u"
                     , !convey->png.nm ? "" : convey->png.nm
                     , cf->num
                     , chctr
                     , opctr
                   ) );
            break;
        case PNGTR_DCHAZ:
        case PNGTR_DBHZ:
        case PNGTR_DBZ:
            PNGTR_O( !!(ex = pngtr_dump(cf, op->val, op->size * op->maxindex, pr))
                   , ( "pngtr_dump %s q%u c%u o%u"
                     , !convey->png.nm ? "" : convey->png.nm
                     , cf->num
                     , chctr
                     , opctr
                   ) );
            break;
        case PNGTR_DBG:
            PNGTR_O( !!(ex = pngtr_dump(cf, op->val, op->size * op->maxindex, pr))
                   , ( "pngtr_dump %s q%u c%u o%u"
                     , !convey->png.nm ? "" : convey->png.nm
                     , cf->num
                     , chctr
                     , opctr
                   ) );
            break;
    }   }
    if  (PNGTR_PRINTEX & f) {
        switch(opdef->op[opctr].otype) {
        case PNGTR_DBZ:
            if  ((!chunk->infla) && !!(ex = pngtr_infla(cf, chctr, opctr))) {
                fflush(pr);
                ifBLIN_QW0( "infla %s q%u c%u o%u"
                          , !convey->png.nm ? "" : convey->png.nm
                          , cf->num
                          , chctr
                          , opctr
                          );
            }
        /* FALLTHRU */
        case PNGTR_DBHZ:
            PNGTR_O( !!(ex = pngtr_dump(cf, chunk->infla, chunk->inflen, pr))
                   , ( "pngtr_dump z %s q%u c%u o%u"
                     , !convey->png.nm ? "" : convey->png.nm
                     , cf->num
                     , chctr
                     , opctr
                   ) );
            break;
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     pr  chctr  cf->num   f                              ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_cprint(pngtr_conf *cf, u_int chctr, BLIN_flag f, FILE *pr) {                               /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey       *convey = &cf->convey[cf->num];
    pngtr_chunk        *chunk  = &convey->chunks[chctr];
    const pngtr_chdef  *chdef;
    const pngtr_chidef *opdef;
    u_int               opnum;
    pngtr_hdr          *hdr;
    int                 ex     = EX_OK;
    u_int32_t           hl;
    u_int8_t           *c      = convey->png.cont;

    ifBLIN_QX3("+ %08X f%08X %08X q%u c%u", cf->flags, f, convey->flags, cf->num, chctr);
    hdr = (pngtr_hdr *)&c[chunk->offset];
    ifBLIN_QX4( "%"BLIN_X" %"BLIN_X" %u %08X"
              , BLIN_I(hdr)
              , BLIN_I(chunk->offset)
              , chunk->csz
              , chunk->flags
              );
    if  ((PNGTR_PRINTHD & f) && !!(ex = chprint(cf, chctr, pr))) goto out;
    if  (4 >= chunk->csz) goto out;
    hl = recobe_load(hdr->ln, -4);
    if  (12 > chunk->csz) goto out;
    ifBLIN_QX5("%u %u", hl, chunk->csz);
    if  ((12 + hl) <= chunk->csz) {
        ifBLIN_QX5( "%08X %08X"
                  , (u_int32_t)recobe_load(CRC(hdr), -4)
                  , recobe_crc32((u_char *)hdr->nm, hl + 4)
                  );
    }
    chdef = &pngtr_chdescr[chunk->type];
    opdef = chdef->ch[(PNGTR_CHCOLOR & chdef->shift) ? convey->IHDR.colortype : 0];
    opnum = opdef->opn;
    if  (PNGTR_PRINTCD & f) {
        PNGTR_O( !!(ex = pngtr_dump(cf, hdr, chunk->csz, pr))
               , ( "pngtr_dump %s %u"
               , !convey->png.nm ? "" : convey->png.nm
               , chctr
               ) );
    }
    for (u_int opctr = 0; opctr < opnum; ++opctr) {
    ;   pngtr_operand *op = &chunk->operands[opctr];

    ;   if  (!op->val) continue;
                                                   /*  pCAL     */
    ;   if  (PNGTR_OPNUM & opdef->op[opctr].var) opnum = 6 + op->val[0];
    ;   if  (!!pngtr_oprint(cf, chctr, opctr, f, pr)) {
    ;       ifBLIN_QW0( "pngtr_oprint %s q%u c%u o%u"
                      , !convey->png.nm ? "" : convey->png.nm
                      , cf->num
                      , chctr
                      , opctr
                      );
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     pr   chctr  cf->num   f                  ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_cvprint(pngtr_conf *cf, u_int chctr, BLIN_flag f, FILE *pr) {                              /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey       *convey = &cf->convey[cf->num];
    pngtr_chunk        *chunk  = &convey->chunks[chctr];
    const pngtr_chdef  *chdef;
    const pngtr_chidef *opdef;
    u_int               opnum;
    int                 ex     = EX_OK;
    int                 l;

    ifBLIN_QX3("+ %08X f%08X %08X q%u c%u", cf->flags, f, convey->flags, cf->num, chctr);
    chdef = &pngtr_chdescr[chunk->type];
    if  (PNGTR_PRINTHD & f) {
        if  (10 > chctr) {
            fprintf(pr, "oooo%1u -------- --------", chctr);
        } else if (100 > chctr) {
            fprintf(pr, "ooo%2u -------- --------", chctr);
        } else if (1000 > chctr) {
            fprintf(pr, "oo%3u -------- --------", chctr);
        } else if (10000 > chctr) {
            fprintf(pr, "o%4u -------- --------", chctr);
        } else {
            fprintf(pr, "%5u -------- --------", chctr);
        }
        l = fprintf( pr
                   , " -- %.*s<%s%s%s%s>"
                   , 4, chdef->nm
                   , !(chdef->nm[0] & 0x20) ? "CRIT" : "anc"
                   , !(chdef->nm[1] & 0x20) ? " pub" : " priv"
                   , !(chdef->nm[2] & 0x20) ? "" : "!"
                   , !(chdef->nm[3] & 0x20) ? "" : " copy"
                   );
        prtl(l + 6, pr);
        fprintf(pr, "%s\n", chdef->descr);
    }
    opdef = chdef->ch[(PNGTR_CHCOLOR & chdef->shift) ? convey->IHDR.colortype : 0];
    opnum = opdef->opn;
    if  (!!chunk->operands) for (u_int opctr = 0; opctr < opnum; ++opctr) {
    ;   pngtr_operand *op = &chunk->operands[opctr];

    ;   if  (!op->val) continue;
                                                   /*  pCAL     */
    ;   if  (PNGTR_OPNUM & opdef->op[opctr].var) opnum = 6 + op->val[0];
    ;   if  (!!pngtr_oprint(cf, chctr, opctr, f, pr)) {
    ;       ifBLIN_QW0( "pngtr_oprint %s q%u c%u o%u"
                      , !convey->png.nm ? "" : convey->png.nm
                      , cf->num
                      , chctr
                      , opctr
                      );
    }   }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     pr  cf->num    f                                      ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_print(pngtr_conf *cf, BLIN_flag f, FILE *pr) {                                             /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey;
    int               ex    = EX_OK;
    u_int8_t         *c;

    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    convey = &cf->convey[cf->num];
    ifBLIN_QX3( "+ %08X f%08X %08X q%u %02X"
              , cf->flags
              , f
              , convey->flags
              , cf->num
              , convey->IHDR.colortype
              );
    if  (PNGTR_PRINTFN & f) {
        u_int64_t type = pngtr_sigs[(convey->IHDR.type > PNGTR_SIGNA_NUM) ? 0 : convey->IHDR.type];

        fprintf( pr
               , "%c%c%c: %s\n"
               , (char)(type >> 48)
               , (char)(type >> 40)
               , (char)(type >> 32)
               , !convey->png.nm ? "" : convey->png.nm
               );
    }
    c = convey->png.cont;
    ifBLIN_QX4("%"BLIN_X, BLIN_I(c));
    for (u_int chctr = 0; chctr < convey->chunum; ++chctr) {
        pngtr_chunk        *chunk;

    ;   chunk = &convey->chunks[chctr];
    ;   if  (!(  ((PNGTR_CHAINSL & convey->flags) && (PNGTR_SELECTD & chunk->flags))
              || ((PNGTR_CHATYPE & convey->flags) && ((1 << chunk->type) & convey->typemask))
              || !((PNGTR_CHAINSL | PNGTR_CHATYPE) & convey->flags)
            ) ) {
    ;       continue;
    ;   }
    ;   if  (PNGTR_VIRTUAL & chunk->flags) {
    ;       if  (!!pngtr_cvprint(cf, chctr, f, pr)) {
    ;           ifBLIN_QW0( "pngtr_cvprint %s q%u c%u"
                          , !convey->png.nm ? "" : convey->png.nm
                          , cf->num
                          , chctr
                          );
    ;       }
    ;   } else if (!!pngtr_cprint(cf, chctr, f, pr)) {
    ;       ifBLIN_QW0( "pngtr_cprint %s q%u c%u"
                      , !convey->png.nm ? "" : convey->png.nm
                      , cf->num
                      , chctr
                      );
    }   }
    if  (  (PNGTR_PRINTHD & f)
        && (  !(PNGTR_CHAINSL & convey->flags)
           || !!(PNGTR_SELECTD & convey->chunks[convey->chunum].flags)
        )  ) {
        fprintf(pr, "      %08X\n", (u_int32_t)convey->png.size);
    }
    if  (PNGTR_PRINTFM & f) {
        PNGTR_O( !!(ex = pngtr_dump(cf, convey->imda, convey->nowsz, pr))
               , ("pngtr_dump img f %s", !convey->png.nm ? "" : convey->png.nm)
               );
    }
    if  ((PNGTR_PRINTFI & f) && !!convey->imda) {
        PNGTR_O( !!(ex = pngtr_flimg(cf, pr))
               , ("pngtr_flimg %s", !convey->png.nm ? "" : convey->png.nm)
               );
        fprintf(pr, "\n");
    }
    if  ((PNGTR_PRINTIM & f) && !!convey->imda) {
        if  (!convey->img.cont) {
            PNGTR_O( !!(ex = pngtr_flimg(cf, NULL))
                   , ("pngtr_flimg %s", !convey->png.nm ? "" : convey->png.nm)
                   );
        }
        PNGTR_O( !!(ex = pngtr_dump(cf, convey->img.cont, convey->img.size, pr))
               , ("pngtr_dump img %s", !convey->png.nm ? "" : convey->png.nm)
               );
    }
out:
    if  (PNGTR_CFTOCH & f) fflush(pr);
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
